#pragma once

#include "ProgramVariables.h"

namespace actions {
namespace expression {

struct ParseContext {
	ProgramVariablesDefinition variables;
};

} // namespace expression
} // namespace actions
