#pragma once

#include "globalincs/pstypes.h"

#include "scripting/ade.h"
#include "scripting/ade_api.h"

namespace scripting::api {

DECLARE_ADE_OBJ(l_TeamColor, int);

} // namespace scripting::api
