#pragma once
/*
 * Created by Mike "MjnMixael" Nelson for the FreeSpace2 Source Code Project.
 * You may not sell or otherwise commercially exploit the source or things you
 * create based on the source.
 *
 * This file is in charge of the "default_settings.tbl", and allows games and mods
 * to define default options settings that will be used on first launch or if a player
 * has not chosen and saved a specific option value.
 */

void default_settings_init();