/*
 * Copyright (C) Volition, Inc. 1999.  All rights reserved.
 *
 * All source code herein is the property of Volition, Inc. You may not sell
 * or otherwise commercially exploit the source or things you created based on the
 * source.
 *
*/



#ifndef _FONT_H
#define _FONT_H

#include "globalincs/pstypes.h"

#include "graphics/software/font.h"
#include "graphics/software/FontManager.h"

#endif
