#pragma once
#ifndef FS2_OPEN_HUDGAUGE_H
#define FS2_OPEN_HUDGAUGE_H

#include "scripting/ade_api.h"
#include "hud/hud.h"

namespace scripting {
namespace api {

DECLARE_ADE_OBJ(l_HudGauge, HudGauge*);

DECLARE_ADE_OBJ(l_HudGaugeDrawFuncs, HudGauge*);

}
}

#endif // FS2_OPEN_HUDGAUGE_H
