// Microsoft Visual C++ generated resource script.
//
#include "freespaceresource.h"

#define APSTUDIO_READONLY_SYMBOLS
/////////////////////////////////////////////////////////////////////////////
//
// Generated from the TEXTINCLUDE 2 resource.
//
#include <windows.h>
/////////////////////////////////////////////////////////////////////////////
#undef APSTUDIO_READONLY_SYMBOLS

/////////////////////////////////////////////////////////////////////////////
// English (United States) resources

#if !defined(AFX_RESOURCE_DLL) || defined(AFX_TARG_ENU)
LANGUAGE LANG_ENGLISH, SUBLANG_ENGLISH_US
#pragma code_page(1252)

#ifdef APSTUDIO_INVOKED
/////////////////////////////////////////////////////////////////////////////
//
// TEXTINCLUDE
//

1 TEXTINCLUDE 
BEGIN
    "freespaceresource.h\0"
END

2 TEXTINCLUDE 
BEGIN
    "#include <windows.h>\0"
END

3 TEXTINCLUDE 
BEGIN
    "#include ""version.rc2""\0"
END

#endif    // APSTUDIO_INVOKED


/////////////////////////////////////////////////////////////////////////////
//
// Dialog
//

DEBUG_MENU DIALOG 0, 0, 185, 159
STYLE DS_SETFONT | DS_MODALFRAME | WS_POPUP | WS_CAPTION | WS_SYSMENU
CAPTION "Debug Menu"
FONT 8, "MS Sans Serif"
BEGIN
    LISTBOX         IDC_LIST_DEBUG,7,7,171,121,LBS_OWNERDRAWFIXED | LBS_NOINTEGRALHEIGHT | WS_VSCROLL | WS_TABSTOP
    PUSHBUTTON      "Ok",IDOK,7,137,171,14
END

FIND_DIALOG DIALOG 0, 0, 119, 47
STYLE DS_SETFONT | DS_MODALFRAME | WS_POPUP | WS_CAPTION | WS_SYSMENU
CAPTION "Find Text"
FONT 8, "MS Sans Serif"
BEGIN
    EDITTEXT        IDC_TEXT,7,7,105,14,ES_AUTOHSCROLL
    DEFPUSHBUTTON   "Find",IDOK,7,26,50,14
    PUSHBUTTON      "Cancel",IDCANCEL,62,26,50,14
END

IDD_CONNECT DIALOG 0, 0, 215, 253
STYLE DS_SETFONT | DS_3DLOOK | WS_POPUP | WS_VISIBLE | WS_CAPTION | WS_SYSMENU
CAPTION "Server"
FONT 8, "MS Sans Serif"
BEGIN
    LTEXT           "# Connections :",IDC_CON_COUNT,7,54,148,10
    LISTBOX         IDC_CONPING,7,79,148,167,WS_VSCROLL | WS_TABSTOP
    LTEXT           "Address and Ping",IDC_STATIC,7,66,143,10
    PUSHBUTTON      "Reset All",IDC_RESET_MULTI,162,174,46,72
    PUSHBUTTON      "Kick",IDC_KICK_BUTTON,161,80,47,19
    CONTROL         "",IDC_STATIC,"Static",SS_GRAYRECT,7,48,201,1
    EDITTEXT        IDC_STD_NAME,64,14,144,14,ES_AUTOHSCROLL
    LTEXT           "Server Name",IDC_STATIC,7,17,42,8
    EDITTEXT        IDC_STD_HOST_PASSWD,64,32,144,14,ES_AUTOHSCROLL
    LTEXT           "Host Password",IDC_STATIC,7,33,48,8
    PUSHBUTTON      "Refresh Missions",IDC_PXO_REFRESH,161,116,47,36,BS_MULTILINE
END

IDD_MULTI DIALOG 0, 0, 215, 253
STYLE DS_SETFONT | DS_MODALFRAME | DS_3DLOOK | WS_POPUP | WS_VISIBLE | WS_CAPTION | WS_SYSMENU
CAPTION "Multi-Player"
FONT 8, "MS Sans Serif"
BEGIN
    LTEXT           "Frame Cap :",10,7,28,43,8
    LTEXT           "",IDC_FRAMECAP_STATIC,53,29,46,10,SS_SUNKEN
    LTEXT           "Realized FPS :",IDC_STATIC_A,119,30,49,12
    LTEXT           "",IDC_STANDALONE_FPS,172,29,34,11,SS_SUNKEN
    LTEXT           "Mission Time :",IDC_STATIC_A2,13,91,49,12
    LTEXT           "",IDC_STANDALONE_MTIME,71,90,125,11,SS_SUNKEN
    CONTROL         "",IDC_STATIC,"Static",SS_WHITERECT | SS_SUNKEN,7,43,199,1
    LTEXT           "Mission Name :",IDC_STATIC,13,80,52,11
    LTEXT           "",IDC_MISSION_NAME,71,78,125,11,SS_SUNKEN
    LTEXT           "Mission Goals",IDC_STATIC,7,140,74,8
    LTEXT           "Netgame Information",IDC_STATIC,133,140,66,8
    LTEXT           "Max Players",IDC_STATIC,117,158,42,8
    LTEXT           "Max Observers",IDC_STATIC,117,184,48,8
    LTEXT           "Security",IDC_STATIC,117,210,26,8
    LTEXT           "Respawns",IDC_STATIC,117,236,34,8
    LTEXT           "",IDC_NG_MAXPLAYERS,178,158,28,8
    LTEXT           "",IDC_NG_MAXOBSERVERS,178,184,28,8
    LTEXT           "",IDC_NG_SECURITY,178,210,28,8
    LTEXT           "",IDC_NG_RESPAWNS,178,236,28,8
END

IDD_PLAYER_DIALOG DIALOG 0, 0, 215, 253
STYLE DS_SETFONT | DS_MODALFRAME | WS_POPUP | WS_CAPTION | WS_SYSMENU
CAPTION "Player Info"
FONT 8, "MS Sans Serif"
BEGIN
    COMBOBOX        IDC_PLAYER_LIST,77,6,112,169,CBS_DROPDOWNLIST | WS_VSCROLL | WS_TABSTOP
    LTEXT           "Ship Type",IDC_STATIC,7,24,38,10
    CONTROL         "",IDC_STATIC,"Static",SS_WHITERECT | SS_NOTIFY | SS_SUNKEN,7,55,201,1
    LTEXT           "",IDC_PSHIP_TYPE,53,24,49,10,SS_SUNKEN
    LTEXT           "Mission Stats",IDC_STATIC,133,59,44,10
    LTEXT           "Primary Shots",IDC_STATIC,7,118,45,9
    LTEXT           "Primary Hits",IDC_STATIC,7,129,40,9
    LTEXT           "Primary BH Hits",IDC_STATIC,7,140,50,9
    LTEXT           "Primary Hit %",IDC_STATIC,7,151,43,9
    LTEXT           "Primary BH Hit %",IDC_STATIC,7,162,55,9
    CONTROL         "",IDC_STATIC,"Static",SS_WHITERECT | SS_NOTIFY | SS_SUNKEN,106,24,1,222
    LTEXT           "All Time Stats",IDC_STATIC,33,59,44,10
    LTEXT           "Primary Shots",IDC_STATIC,109,118,45,9
    LTEXT           "Primary Hits",IDC_STATIC,109,129,40,9
    LTEXT           "Primary BH Hits",IDC_STATIC,109,140,50,9
    LTEXT           "Primary Hit %",IDC_STATIC,109,151,43,9
    LTEXT           "Primary BH Hit %",IDC_STATIC,109,162,55,9
    LTEXT           "Secondary Shots",IDC_STATIC,7,173,56,9
    LTEXT           "Secondary Hits",IDC_STATIC,7,184,49,9
    LTEXT           "Secondary BH Hits",IDC_STATIC,7,195,63,9
    LTEXT           "Secondary Hit %",IDC_STATIC,7,206,55,9
    LTEXT           "Secondary BH Hit %",IDC_STATIC,7,217,65,9
    LTEXT           "Secondary Shots",IDC_STATIC,109,173,56,9
    LTEXT           "Secondary Hits",IDC_STATIC,109,184,49,9
    LTEXT           "Secondary BH Hits",IDC_STATIC,109,195,63,9
    LTEXT           "Secondary Hit %",IDC_STATIC,109,206,55,9
    LTEXT           "Secondary BH Hit %",IDC_STATIC,109,217,65,9
    LTEXT           "",IDC_PSHOTS,75,118,28,8,SS_SUNKEN
    LTEXT           "",IDC_PHITS,75,129,28,8,SS_SUNKEN
    LTEXT           "",IDC_PBHHITS,75,140,28,8,SS_SUNKEN
    LTEXT           "",IDC_PPCT,75,151,28,8,SS_SUNKEN
    LTEXT           "",IDC_PBHPCT,75,162,28,8,SS_SUNKEN
    LTEXT           "",IDC_SSHOTS,75,173,28,8,SS_SUNKEN
    LTEXT           "",IDC_SECHITS,75,184,28,8,SS_SUNKEN
    LTEXT           "",IDC_SBHHITS,75,195,28,8,SS_SUNKEN
    LTEXT           "",IDC_SPCT,75,206,28,8,SS_SUNKEN
    LTEXT           "",IDC_SBHPCT,75,217,28,8,SS_SUNKEN
    LTEXT           "",IDC_MPSHOTS,180,118,28,8,SS_SUNKEN
    LTEXT           "",IDC_MPHITS,180,129,28,8,SS_SUNKEN
    LTEXT           "",IDC_MPBHHITS,180,140,28,8,SS_SUNKEN
    LTEXT           "",IDC_MPPCT,180,151,28,8,SS_SUNKEN
    LTEXT           "",IDC_MPBHPCT,180,162,28,9,SS_SUNKEN
    LTEXT           "",IDC_MSSHOTS,180,173,28,8,SS_SUNKEN
    LTEXT           "",IDC_MSECHITS,180,184,28,8,SS_SUNKEN
    LTEXT           "",IDC_MSBHHITS,180,195,28,8,SS_SUNKEN
    LTEXT           "",IDC_MSPCT,180,206,28,8,SS_SUNKEN
    LTEXT           "",IDC_MSBHPCT,180,217,28,8,SS_SUNKEN
    CONTROL         "",IDC_PINFO_PING,"Static",SS_WHITERECT | SS_NOTIFY | SS_SUNKEN,7,235,201,1
    LTEXT           "Player",IDC_STATIC,44,6,20,8
    LTEXT           "",IDC_PING_TIME,53,38,49,10,SS_SUNKEN
    LTEXT           "Avg Ping",IDC_STATIC,7,38,32,9
    LTEXT           "Assists",IDC_STATIC,7,107,54,10
    LTEXT           "Assists",IDC_STATIC,109,107,54,10
    LTEXT           "",IDC_ASSISTS,75,107,28,8,SS_SUNKEN
    LTEXT           "",IDC_MASSISTS,180,107,28,8,SS_SUNKEN
    LTEXT           "Score",IDC_STATIC,7,73,20,8
    LTEXT           "Score",IDC_STATIC,109,73,20,8
    LTEXT           "Kills",IDC_STATIC,7,84,13,8
    LTEXT           "Kills",IDC_STATIC,109,84,13,8
    LTEXT           "Friendly Kills",IDC_STATIC,109,96,39,8
    LTEXT           "Friendly Kills",IDC_STATIC,7,96,39,8
    LTEXT           "",IDC_SCORE,75,74,28,8,SS_SUNKEN
    LTEXT           "",IDC_KILL_COUNT,75,85,28,8,SS_SUNKEN
    LTEXT           "",IDC_KILL_COUNT_BH,75,96,28,8,SS_SUNKEN
    LTEXT           "",IDC_MSCORE,180,74,28,8,SS_SUNKEN
    LTEXT           "",IDC_MKILL_COUNT,180,85,28,8,SS_SUNKEN
    LTEXT           "",IDC_MKILL_COUNT_BH,180,96,28,8,SS_SUNKEN
END

IDD_GODSTUFF DIALOG 0, 0, 215, 253
STYLE DS_SETFONT | DS_MODALFRAME | WS_POPUP | WS_CAPTION | WS_SYSMENU
CAPTION "God Stuff"
FONT 8, "MS Sans Serif"
BEGIN
    COMBOBOX        IDC_PLAYER_GOD_LIST,39,7,101,211,CBS_DROPDOWNLIST | WS_VSCROLL | WS_TABSTOP
    LTEXT           "",IDC_GODSTUFF_FPS,177,7,31,8,SS_SUNKEN
    LTEXT           "FPS",IDC_STATIC,154,7,16,8
    LTEXT           "Player ",IDC_STATIC,7,7,20,12
    EDITTEXT        IDC_GODSTUFF_BROADCAST,7,47,201,12,ES_MULTILINE
    LTEXT           "Server Message",IDC_STATIC,7,39,54,8
    PUSHBUTTON      "Send",IDC_GODSTUFF_SENDMESS,7,59,70,13
    LISTBOX         IDC_GOD_CHAT,7,84,201,162,LBS_NOINTEGRALHEIGHT | LBS_NOSEL | WS_VSCROLL | WS_HSCROLL | WS_TABSTOP
END

IDD_DEBUG_DIALOG DIALOG 0, 0, 215, 253
STYLE DS_SETFONT | DS_MODALFRAME | WS_POPUP | WS_CAPTION | WS_SYSMENU
CAPTION "Debug"
FONT 8, "MS Sans Serif"
BEGIN
    LTEXT           "Standalone State :",IDC_STATIC,7,7,61,11
    LTEXT           "",IDC_STANDALONE_STATE,71,7,125,11,SS_SUNKEN
    LTEXT           "Multi log:",IDC_STATIC,8,26,29,8
    LISTBOX         IDC_MULTILOG,7,37,201,209,NOT LBS_NOTIFY | LBS_NOINTEGRALHEIGHT | LBS_NOSEL | WS_VSCROLL | WS_HSCROLL | WS_TABSTOP
END

IDD_GEN DIALOG 0, 0, 186, 95
STYLE DS_SETFONT | DS_MODALFRAME | DS_SETFOREGROUND | DS_3DLOOK | WS_POPUP | WS_VISIBLE | WS_CAPTION | WS_SYSMENU
CAPTION "Dialog"
FONT 8, "MS Sans Serif"
BEGIN
    CTEXT           "",IDC_FIELD1,27,11,136,8
    CTEXT           "",IDC_FIELD2,28,30,130,8
    CTEXT           "",IDC_FIELD3,27,49,138,8
END

IDD_DIALOG1 DIALOG 0, 0, 186, 95
STYLE DS_SETFONT | DS_MODALFRAME | WS_POPUP | WS_CAPTION | WS_SYSMENU
CAPTION "Dialog"
FONT 8, "MS Sans Serif"
BEGIN
    DEFPUSHBUTTON   "OK",IDOK,129,7,50,14
    PUSHBUTTON      "Cancel",IDCANCEL,129,24,50,14
END


/////////////////////////////////////////////////////////////////////////////
//
// DESIGNINFO
//

#ifdef APSTUDIO_INVOKED
GUIDELINES DESIGNINFO
BEGIN
    "DEBUG_MENU", DIALOG
    BEGIN
        LEFTMARGIN, 7
        RIGHTMARGIN, 178
        TOPMARGIN, 7
        BOTTOMMARGIN, 152
    END

    "FIND_DIALOG", DIALOG
    BEGIN
        LEFTMARGIN, 7
        RIGHTMARGIN, 112
        TOPMARGIN, 7
        BOTTOMMARGIN, 40
    END

    IDD_CONNECT, DIALOG
    BEGIN
        LEFTMARGIN, 7
        RIGHTMARGIN, 208
        TOPMARGIN, 7
        BOTTOMMARGIN, 246
    END

    IDD_MULTI, DIALOG
    BEGIN
        LEFTMARGIN, 7
        RIGHTMARGIN, 206
        TOPMARGIN, 7
        BOTTOMMARGIN, 246
    END

    IDD_PLAYER_DIALOG, DIALOG
    BEGIN
        LEFTMARGIN, 7
        RIGHTMARGIN, 208
        TOPMARGIN, 6
        BOTTOMMARGIN, 246
    END

    IDD_GODSTUFF, DIALOG
    BEGIN
        LEFTMARGIN, 7
        RIGHTMARGIN, 208
        TOPMARGIN, 7
        BOTTOMMARGIN, 246
    END

    IDD_DEBUG_DIALOG, DIALOG
    BEGIN
        LEFTMARGIN, 7
        RIGHTMARGIN, 208
        TOPMARGIN, 7
        BOTTOMMARGIN, 246
    END

    IDD_GEN, DIALOG
    BEGIN
        LEFTMARGIN, 7
        RIGHTMARGIN, 179
        TOPMARGIN, 5
        BOTTOMMARGIN, 88
    END

    IDD_DIALOG1, DIALOG
    BEGIN
        LEFTMARGIN, 7
        RIGHTMARGIN, 179
        TOPMARGIN, 7
        BOTTOMMARGIN, 88
    END
END
#endif    // APSTUDIO_INVOKED

/////////////////////////////////////////////////////////////////////////////
//
// Bitmap
//

IDB_GOAL_INC            BITMAP  DISCARDABLE     "goal_inc.bmp"
IDB_GOAL_COMP           BITMAP  DISCARDABLE     "goal_com.bmp"
IDB_GOAL_ORD            BITMAP  DISCARDABLE     "goal_ord.bmp"
IDB_GOAL_NONE           BITMAP  DISCARDABLE     "goal_none.bmp"
IDB_GOAL_FAIL           BITMAP  DISCARDABLE     "goal_fail.bmp"

/////////////////////////////////////////////////////////////////////////////
//
// Icon
//
// Icon with lowest ID value placed first to ensure application icon
// remains consistent on all systems.
/* VS2008 can support stuff that these builds can't */

#ifndef _VC08
 /* Non-VS2008 builds get one of two icons */
 #ifdef _DEBUG
  IDI_APP_ICON            ICON    DISCARDABLE     "dbg_icon.ico"
 #else
  IDI_APP_ICON            ICON    DISCARDABLE     "app_icon.ico"
 #endif
#else
 /* VS2008 builds get cool icons */
 #ifdef _DEBUG
  #ifdef _SSE2
   IDI_APP_ICON            ICON    DISCARDABLE     "V_sse-d.ico"
  #else
   IDI_APP_ICON            ICON    DISCARDABLE     "V_debug.ico"
  #endif
 #else
  #ifdef _SSE2
   IDI_APP_ICON            ICON    DISCARDABLE     "V_sse.ico"
  #else
   IDI_APP_ICON            ICON    DISCARDABLE     "V_app.ico"
  #endif
 #endif
#endif

#endif    // English (U.S.) resources
/////////////////////////////////////////////////////////////////////////////


/////////////////////////////////////////////////////////////////////////////
// English (U.K.) resources

#if !defined(AFX_RESOURCE_DLL) || defined(AFX_TARG_ENG)
#ifdef _WIN32
LANGUAGE LANG_ENGLISH, SUBLANG_ENGLISH_UK
#pragma code_page(1252)
#endif //_WIN32

#endif    // English (U.K.) resources
/////////////////////////////////////////////////////////////////////////////


#ifndef APSTUDIO_INVOKED
/////////////////////////////////////////////////////////////////////////////
//
// Generated from the TEXTINCLUDE 3 resource.
//
#include "version.rc2"

/////////////////////////////////////////////////////////////////////////////
#endif    // not APSTUDIO_INVOKED


/////////////////////////////////////////////////////////////////////////////
//
// SRGRAMMAR
//

#ifdef FS2_VOICER
IDR_CMD_CFG             SRGRAMMAR DISCARDABLE   "code\sound\phrases.cfg"
#endif
