#pragma once

#include "scripting/ade_api.h"
#include "scripting/api/LuaPromise.h"

namespace scripting {
namespace api {

DECLARE_ADE_OBJ(l_Promise, LuaPromise);

} // namespace api
} // namespace scripting
