#pragma once

#include "scripting/ade_api.h"

namespace scripting {
namespace api {

//**********HANDLE: Persona
DECLARE_ADE_OBJ(l_Persona, int);


//**********HANDLE: Message
DECLARE_ADE_OBJ(l_Message, int);

}
}

